﻿using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Analysis.FitnessLandscape.DataTables;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Analysis.FitnessLandscape.Algorithms {

  [Item("EvolvabilityAnalyzer", "Analyzes and consolidates repeated (very short) trails to evolvability metrics.")]
  [StorableClass]
  public class EvolvabilityAnalyzer : SingleSuccessorOperator, IRepeatsAnalyzer {

    #region Parameters
    public ScopeTreeLookupParameter<DoubleValue> FinalQualitiesParameter {
      get { return (ScopeTreeLookupParameter<DoubleValue>)Parameters["FinalQualities"]; }
    }
    public LookupParameter<DoubleValue> BaseQualityParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["BaseQuality"]; }
    }
    public LookupParameter<DoubleValue> E_a_Parameter {
      get { return (LookupParameter<DoubleValue>)Parameters["E_a"]; }
    }
    public LookupParameter<DoubleValue> E_b_Parameter {
      get { return (LookupParameter<DoubleValue>)Parameters["E_b"]; }
    }
    public LookupParameter<DoubleValue> E_c_Parameter {
      get { return (LookupParameter<DoubleValue>)Parameters["E_c"]; }
    }
    public LookupParameter<DoubleValue> E_d_Parameter {
      get { return (LookupParameter<DoubleValue>)Parameters["E_d"]; }
    }
    public LookupParameter<DoubleValue> ExpectedAbsoluteQualityDeltaParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["ExpectedAbsoluteQualityDelta"]; }
    }
    public LookupParameter<DoubleValue> ExpectedRelativeQualityDeltaParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["ExpectedRelativeQualityDelta"]; }
    }    
    public LookupParameter<ResultCollection> ResultsParameter {
      get { return (LookupParameter<ResultCollection>)Parameters["Results"]; }
    }
    #endregion

    #region Construction & Cloning
    [StorableConstructor]
    protected EvolvabilityAnalyzer(bool deserializing) : base(deserializing) { }
    protected EvolvabilityAnalyzer(EvolvabilityAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public EvolvabilityAnalyzer() {
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("FinalQualities", "Quality values of the end of the trails."));
      Parameters.Add(new LookupParameter<DoubleValue>("BaseQuality", "Initial quality value."));

      Parameters.Add(new LookupParameter<DoubleValue>("E_a", "Probability of non-deleterious mutation."));
      Parameters.Add(new LookupParameter<DoubleValue>("E_b", "Average expected offspring fitness."));
      Parameters.Add(new LookupParameter<DoubleValue>("E_c", "90th percentile offspring fitness."));
      Parameters.Add(new LookupParameter<DoubleValue>("E_d", "10th percentile offspring fitness."));
      Parameters.Add(new LookupParameter<DoubleValue>("ExpectedAbsoluteQualityDelta", "Expected absolute quality delta."));
      Parameters.Add(new LookupParameter<DoubleValue>("ExpectedRelativeQualityDelta", "Expected relative quality delta."));

      Parameters.Add(new LookupParameter<ResultCollection>("Results", "The collection of all results of this algorithm"));
      FinalQualitiesParameter.Depth = 2;
      FinalQualitiesParameter.ActualName = "Quality";
      BaseQualityParameter.ActualName = "Quality";      
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new EvolvabilityAnalyzer(this, cloner);
    }
    #endregion

    public override IOperation Apply() {
      double baseQuality = BaseQualityParameter.ActualValue.Value;
      var finalQualities = FinalQualitiesParameter.ActualValue.Select(v => v.Value).ToList();
      E_a_Parameter.ActualValue = new DoubleValue(1.0 * finalQualities.Count(fq => fq >= baseQuality) / finalQualities.Count);
      if (finalQualities.Count > 0) {
        DistributionAnalyzer a = new DistributionAnalyzer(finalQualities);
        double mean = a.Mean;
        E_b_Parameter.ActualValue = new DoubleValue(mean);
        E_c_Parameter.ActualValue = new DoubleValue(a[0.9]);
        E_d_Parameter.ActualValue = new DoubleValue(a[0.1]);
        ExpectedAbsoluteQualityDeltaParameter.ActualValue = new DoubleValue(mean - baseQuality);
        ExpectedRelativeQualityDeltaParameter.ActualValue = new DoubleValue((mean - baseQuality)/baseQuality);
        var results = ResultsParameter.ActualValue;
        results.Add(new Result("Base Quality", (DoubleValue)BaseQualityParameter.ActualValue.Clone()));
        results.Add(new Result("E_a", (DoubleValue)E_a_Parameter.ActualValue.Clone()));
        results.Add(new Result("E_b", (DoubleValue)E_b_Parameter.ActualValue.Clone()));
        results.Add(new Result("E_c", (DoubleValue)E_c_Parameter.ActualValue.Clone()));
        results.Add(new Result("E_d", (DoubleValue)E_d_Parameter.ActualValue.Clone()));
        results.Add(new Result("Expected Absolute Quality Delta", (DoubleValue)ExpectedAbsoluteQualityDeltaParameter.ActualValue.Clone()));
        results.Add(new Result("Expected Relative Quality Delta", (DoubleValue)ExpectedRelativeQualityDeltaParameter.ActualValue.Clone()));
      }      
      return base.Apply();
    }
  }
}
